from __future__ import annotations

from logging import Filter, LogRecord, getLogger
from re import search
from typing import TYPE_CHECKING, Any, ClassVar, override

from pydantic_settings.sources import DEFAULT_PATH
from pydantic_settings_sops import SOPSConfigSettingsSource

from utilities.pydantic_settings import (
    CustomBaseSettings,
    PathLikeOrWithSection,
    _ensure_section,
    _get_section,
)

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from pydantic_settings import BaseSettings, PydanticBaseSettingsSource
    from pydantic_settings.sources import PathType

    from utilities.types import MaybeSequenceStr


class _SuppressDefaultConfigMessage(Filter):
    @override
    def filter(self, record: LogRecord) -> bool:
        return not search(
            r"^default config file does not exists '.*'$", record.getMessage()
        )


getLogger("sopsy.utils").addFilter(_SuppressDefaultConfigMessage())


class SopsBaseSettings(CustomBaseSettings):
    """Base settings for loading secrets using `sops/age`."""

    # paths
    secret_files: ClassVar[Sequence[PathLikeOrWithSection]] = []

    @classmethod
    @override
    def _yield_base_settings_sources(
        cls,
        settings_cls: type[BaseSettings],
        env_settings: PydanticBaseSettingsSource,
        /,
    ) -> Iterator[PydanticBaseSettingsSource]:
        yield from super()._yield_base_settings_sources(settings_cls, env_settings)
        for file, section in map(_ensure_section, cls.secret_files):
            yield SOPSConfigSectionSettingsSource(
                settings_cls, json_file=file, section=section
            )


class SOPSConfigSectionSettingsSource(SOPSConfigSettingsSource):
    @override
    def __init__(
        self,
        settings_cls: type[BaseSettings],
        json_file: PathType | None = DEFAULT_PATH,
        yaml_file: PathType | None = DEFAULT_PATH,
        *,
        section: MaybeSequenceStr,
    ) -> None:
        super().__init__(settings_cls, json_file=json_file, yaml_file=yaml_file)  # pyright: ignore[reportArgumentType]
        self.section = section

    @override
    def __call__(self) -> dict[str, Any]:
        return _get_section(super().__call__(), self.section)


__all__ = ["SOPSConfigSectionSettingsSource", "SopsBaseSettings"]



from __future__ import annotations

import tempfile
from contextlib import contextmanager
from pathlib import Path
from shutil import move
from tempfile import NamedTemporaryFile as _NamedTemporaryFile
from tempfile import gettempdir as _gettempdir
from typing import TYPE_CHECKING, override

from utilities.warnings import suppress_warnings

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import TracebackType

    from utilities.types import PathLike


class TemporaryDirectory:
    """Wrapper around `TemporaryDirectory` with a `Path` attribute."""

    def __init__(
        self,
        *,
        suffix: str | None = None,
        prefix: str | None = None,
        dir: PathLike | None = None,  # noqa: A002
        ignore_cleanup_errors: bool = False,
        delete: bool = True,
    ) -> None:
        super().__init__()
        self._temp_dir = _TemporaryDirectoryNoResourceWarning(
            suffix=suffix,
            prefix=prefix,
            dir=dir,
            ignore_cleanup_errors=ignore_cleanup_errors,
            delete=delete,
        )
        self.path = Path(self._temp_dir.name)

    def __enter__(self) -> Path:
        return Path(self._temp_dir.__enter__())

    def __exit__(
        self,
        exc: type[BaseException] | None,
        val: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self._temp_dir.__exit__(exc, val, tb)


class _TemporaryDirectoryNoResourceWarning(tempfile.TemporaryDirectory):
    @classmethod
    @override
    def _cleanup(  # pyright: ignore[reportGeneralTypeIssues]
        cls,
        name: str,
        warn_message: str,
        ignore_errors: bool = False,
        delete: bool = True,
    ) -> None:
        with suppress_warnings(category=ResourceWarning):
            return super()._cleanup(  # pyright: ignore[reportAttributeAccessIssue]
                name, warn_message, ignore_errors=ignore_errors, delete=delete
            )


##


@contextmanager
def TemporaryFile(  # noqa: N802
    *,
    suffix: str | None = None,
    prefix: str | None = None,
    dir: PathLike | None = None,  # noqa: A002
    ignore_cleanup_errors: bool = False,
    delete: bool = True,
    name: str | None = None,
) -> Iterator[Path]:
    """Yield a temporary file."""
    with TemporaryDirectory(
        suffix=suffix,
        prefix=prefix,
        dir=dir,
        ignore_cleanup_errors=ignore_cleanup_errors,
        delete=delete,
    ) as temp_dir:
        temp_file = _NamedTemporaryFile(  # noqa: SIM115
            dir=temp_dir, delete=delete, delete_on_close=False
        )
        if name is None:
            yield temp_dir / temp_file.name
        else:
            _ = move(temp_dir / temp_file.name, temp_dir / name)
            yield temp_dir / name


##


def gettempdir() -> Path:
    """Get the name of the directory used for temporary files."""
    return Path(_gettempdir())


TEMP_DIR = gettempdir()


__all__ = ["TEMP_DIR", "TemporaryDirectory", "TemporaryFile", "gettempdir"]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from pytest import FixtureRequest

    from utilities.pytest_regressions import (
        OrjsonRegressionFixture,
        PolarsRegressionFixture,
    )


try:
    from pytest import fixture
except ModuleNotFoundError:
    pass
else:

    @fixture
    def orjson_regression(
        *, request: FixtureRequest, tmp_path: Path
    ) -> OrjsonRegressionFixture:
        """Instance of the `OrjsonRegressionFixture`."""
        from utilities.pytest_regressions import OrjsonRegressionFixture

        path = _get_path(request)
        return OrjsonRegressionFixture(path, request, tmp_path)

    @fixture
    def polars_regression(
        *, request: FixtureRequest, tmp_path: Path
    ) -> PolarsRegressionFixture:
        """Instance of the `PolarsRegressionFixture`."""
        from utilities.pytest_regressions import PolarsRegressionFixture

        path = _get_path(request)
        return PolarsRegressionFixture(path, request, tmp_path)


def _get_path(request: FixtureRequest, /) -> Path:
    from utilities.pathlib import get_root
    from utilities.pytest import _NodeIdToPathNotGetTailError, node_id_path

    path = Path(cast("Any", request).fspath)
    root = Path("src", "tests")
    try:
        tail = node_id_path(request.node.nodeid, root=root)
    except _NodeIdToPathNotGetTailError:
        root = Path("tests")
        tail = node_id_path(request.node.nodeid, root=root)
    return get_root(path).joinpath(root, "regressions", tail)


__all__ = ["orjson_regression", "polars_regression"]



from __future__ import annotations

from pathlib import Path
from typing import Annotated

from pydantic import BeforeValidator

ExpandedPath = Annotated[Path, BeforeValidator(lambda p: Path(p).expanduser())]


__all__ = ["ExpandedPath"]



from __future__ import annotations

import datetime as dt
import decimal
from contextlib import suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast, overload, override
from uuid import UUID

import polars as pl
from polars import (
    Binary,
    DataFrame,
    Date,
    Datetime,
    Float64,
    Int32,
    Int64,
    String,
    Time,
    UInt32,
    UInt64,
    concat,
    read_database,
)
from sqlalchemy import Column, Select, select
from sqlalchemy.exc import DuplicateColumnError

from utilities.asyncio import timeout_td
from utilities.functions import identity
from utilities.iterables import (
    CheckDuplicatesError,
    OneError,
    check_duplicates,
    chunked,
    one,
)
from utilities.polars import zoned_date_time_dtype
from utilities.reprlib import get_repr
from utilities.sqlalchemy import (
    CHUNK_SIZE_FRAC,
    TableOrORMInstOrClass,
    ensure_tables_created,
    get_chunk_size,
    get_columns,
    insert_items,
)
from utilities.text import snake_case
from utilities.typing import is_subclass_gen
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterable,
        AsyncIterator,
        Iterable,
        Iterator,
        Mapping,
    )

    from polars._typing import PolarsDataType, SchemaDict
    from sqlalchemy.ext.asyncio import AsyncEngine
    from sqlalchemy.sql import ColumnCollection
    from sqlalchemy.sql.base import ReadOnlyColumnCollection
    from whenever import TimeDelta

    from utilities.types import Delta, MaybeType, TimeZoneLike


async def insert_dataframe(
    df: DataFrame,
    table_or_orm: TableOrORMInstOrClass,
    engine: AsyncEngine,
    /,
    *,
    snake: bool = False,
    is_upsert: bool = False,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    timeout_create: TimeDelta | None = None,
    error_create: type[Exception] = TimeoutError,
    timeout_insert: TimeDelta | None = None,
    error_insert: type[Exception] = TimeoutError,
) -> None:
    """Insert/upsert a DataFrame into a database."""
    mapping = _insert_dataframe_map_df_schema_to_table(
        df.schema, table_or_orm, snake=snake
    )
    items = df.select(*mapping).rename(mapping).rows(named=True)
    if len(items) == 0:
        if not assume_tables_exist:
            await ensure_tables_created(
                engine, table_or_orm, timeout=timeout_create, error=error_create
            )
        return
    await insert_items(
        engine,
        (items, table_or_orm),
        snake=snake,
        is_upsert=is_upsert,
        chunk_size_frac=chunk_size_frac,
        assume_tables_exist=assume_tables_exist,
        timeout_create=timeout_create,
        error_create=error_create,
        timeout_insert=timeout_insert,
        error_insert=error_insert,
    )


def _insert_dataframe_map_df_schema_to_table(
    df_schema: SchemaDict,
    table_or_orm: TableOrORMInstOrClass,
    /,
    *,
    snake: bool = False,
) -> dict[str, str]:
    """Map a DataFrame schema to a table."""
    table_schema = {col.name: col.type.python_type for col in get_columns(table_or_orm)}
    out: dict[str, str] = {}
    for df_col_name, df_col_type in df_schema.items():
        with suppress(_InsertDataFrameMapDFColumnToTableColumnAndTypeError):
            out[df_col_name] = _insert_dataframe_map_df_column_to_table_schema(
                df_col_name, df_col_type, table_schema, snake=snake
            )
    return out


def _insert_dataframe_map_df_column_to_table_schema(
    df_col_name: str,
    df_col_type: PolarsDataType,
    table_schema: Mapping[str, type[Any]],
    /,
    *,
    snake: bool = False,
) -> str:
    """Map a DataFrame column to a table schema."""
    db_col_name, db_col_type = _insert_dataframe_map_df_column_to_table_column_and_type(
        df_col_name, table_schema, snake=snake
    )
    if not _insert_dataframe_check_df_and_db_types(df_col_type, db_col_type):
        msg = f"{df_col_type=}, {db_col_type=}"
        raise _InsertDataFrameMapDFColumnToTableSchemaError(msg)
    return db_col_name


class _InsertDataFrameMapDFColumnToTableSchemaError(Exception): ...


def _insert_dataframe_map_df_column_to_table_column_and_type(
    df_col_name: str, table_schema: Mapping[str, type[Any]], /, *, snake: bool = False
) -> tuple[str, type[Any]]:
    """Map a DataFrame column to a table column and type."""
    items = table_schema.items()
    func = snake_case if snake else identity
    target = func(df_col_name)
    try:
        return one((n, t) for n, t in items if func(n) == target)
    except OneError:
        raise _InsertDataFrameMapDFColumnToTableColumnAndTypeError(
            df_col_name=df_col_name, table_schema=table_schema, snake=snake
        ) from None


@dataclass(kw_only=True, slots=True)
class _InsertDataFrameMapDFColumnToTableColumnAndTypeError(Exception):
    df_col_name: str
    table_schema: Mapping[str, type[Any]]
    snake: bool

    @override
    def __str__(self) -> str:
        return f"Unable to map DataFrame column {self.df_col_name!r} into table schema {get_repr(self.table_schema)} with snake={self.snake}"


def _insert_dataframe_check_df_and_db_types(
    dtype: PolarsDataType, db_col_type: type, /
) -> bool:
    return (
        ((dtype == pl.Boolean) and is_subclass_gen(db_col_type, bool))
        or ((dtype == Date) and is_subclass_gen(db_col_type, dt.date))
        or ((dtype == Datetime) and is_subclass_gen(db_col_type, dt.datetime))
        or ((dtype == Float64) and issubclass(db_col_type, float))
        or ((dtype == Int32) and is_subclass_gen(db_col_type, int))
        or ((dtype == Int64) and is_subclass_gen(db_col_type, int))
        or ((dtype == UInt32) and is_subclass_gen(db_col_type, int))
        or ((dtype == UInt64) and is_subclass_gen(db_col_type, int))
        or ((dtype == String) and issubclass(db_col_type, str))
    )


@overload
async def select_to_dataframe(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = False,
    time_zone: TimeZoneLike = UTC,
    batch_size: None = None,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = None,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    **kwargs: Any,
) -> DataFrame: ...
@overload
async def select_to_dataframe(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = False,
    time_zone: TimeZoneLike = UTC,
    batch_size: int,
    in_clauses: None = None,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    **kwargs: Any,
) -> Iterable[DataFrame]: ...
@overload
async def select_to_dataframe(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = False,
    time_zone: TimeZoneLike = UTC,
    batch_size: int,
    in_clauses: tuple[Column[Any], Iterable[Any]],
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    **kwargs: Any,
) -> AsyncIterable[DataFrame]: ...
@overload
async def select_to_dataframe(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = False,
    time_zone: TimeZoneLike = UTC,
    batch_size: int | None = None,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = None,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    **kwargs: Any,
) -> DataFrame | Iterable[DataFrame] | AsyncIterable[DataFrame]: ...
async def select_to_dataframe(
    sel: Select[Any],
    engine: AsyncEngine,
    /,
    *,
    snake: bool = False,
    time_zone: TimeZoneLike = UTC,
    batch_size: int | None = None,
    in_clauses: tuple[Column[Any], Iterable[Any]] | None = None,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    timeout: Delta | None = None,
    error: MaybeType[BaseException] = TimeoutError,
    **kwargs: Any,
) -> DataFrame | Iterable[DataFrame] | AsyncIterable[DataFrame]:
    """Read a table from a database into a DataFrame."""
    if snake:
        sel = _select_to_dataframe_apply_snake(sel)
    schema = _select_to_dataframe_map_select_to_df_schema(sel, time_zone=time_zone)
    if in_clauses is None:
        async with timeout_td(timeout, error=error):
            return read_database(
                sel,
                cast("Any", engine),
                iter_batches=batch_size is not None,
                batch_size=batch_size,
                schema_overrides=schema,
                **kwargs,
            )
    sels = _select_to_dataframe_yield_selects_with_in_clauses(
        sel,
        engine,
        in_clauses,
        in_clauses_chunk_size=in_clauses_chunk_size,
        chunk_size_frac=chunk_size_frac,
    )
    if batch_size is None:
        async with timeout_td(timeout, error=error):
            dfs = [
                await select_to_dataframe(
                    sel,
                    engine,
                    snake=snake,
                    time_zone=time_zone,
                    batch_size=None,
                    in_clauses=None,
                    timeout=timeout,
                    error=error,
                    **kwargs,
                )
                for sel in sels
            ]
        try:
            return concat(dfs)
        except ValueError:
            return DataFrame(schema=schema)

    async def yield_dfs() -> AsyncIterator[DataFrame]:
        async with timeout_td(timeout, error=error):
            for sel_i in sels:
                for df in await select_to_dataframe(
                    sel_i,
                    engine,
                    snake=snake,
                    time_zone=time_zone,
                    batch_size=batch_size,
                    in_clauses=None,
                    chunk_size_frac=chunk_size_frac,
                    timeout=timeout,
                    error=error,
                    **kwargs,
                ):
                    yield df

    return yield_dfs()


def _select_to_dataframe_apply_snake(sel: Select[Any], /) -> Select[Any]:
    """Apply snake-case to a selectable."""
    alias = sel.alias()
    columns = [alias.c[c.name].label(snake_case(c.name)) for c in sel.selected_columns]
    return select(*columns)


def _select_to_dataframe_map_select_to_df_schema(
    sel: Select[Any], /, *, time_zone: TimeZoneLike = UTC
) -> SchemaDict:
    """Map a select to a DataFrame schema."""
    columns: ReadOnlyColumnCollection = cast("Any", sel).selected_columns
    _select_to_dataframe_check_duplicates(columns)
    return {
        col.name: _select_to_dataframe_map_table_column_type_to_dtype(
            col.type, time_zone=time_zone
        )
        for col in columns
    }


def _select_to_dataframe_map_table_column_type_to_dtype(
    type_: Any, /, *, time_zone: TimeZoneLike = UTC
) -> PolarsDataType:
    """Map a table column type to a polars type."""
    type_use = type_() if isinstance(type_, type) else type_
    py_type = type_use.python_type
    if is_subclass_gen(py_type, bool):
        return pl.Boolean
    if issubclass(py_type, bytes):
        return Binary
    if issubclass(py_type, decimal.Decimal):
        return pl.Decimal
    if is_subclass_gen(py_type, dt.date):
        return pl.Date
    if is_subclass_gen(py_type, dt.datetime):
        has_tz: bool = type_use.timezone
        return zoned_date_time_dtype(time_zone=time_zone) if has_tz else Datetime()
    if issubclass(py_type, dt.time):
        return Time
    if issubclass(py_type, dt.timedelta):
        return pl.Duration
    if issubclass(py_type, float):
        return Float64
    if is_subclass_gen(py_type, int):
        return Int64
    if issubclass(py_type, UUID | str):
        return String
    msg = f"{type_=}, {py_type=}"  # pragma: no cover
    raise _SelectToDataFrameMapTableColumnToDTypeError(msg)  # pragma: no cover


class _SelectToDataFrameMapTableColumnToDTypeError(Exception): ...


def _select_to_dataframe_check_duplicates(
    columns: ColumnCollection[Any, Any], /
) -> None:
    """Check a select for duplicate columns."""
    names = [col.name for col in columns]
    try:
        check_duplicates(names)
    except CheckDuplicatesError as error:
        msg = f"Columns must not contain duplicates; got {error.counts}"
        raise DuplicateColumnError(msg) from None


def _select_to_dataframe_yield_selects_with_in_clauses(
    sel: Select[Any],
    engine: AsyncEngine,
    in_clauses: tuple[Column[Any], Iterable[Any]],
    /,
    *,
    in_clauses_chunk_size: int | None = None,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> Iterator[Select[Any]]:
    in_col, in_values = in_clauses
    if in_clauses_chunk_size is None:
        chunk_size = get_chunk_size(
            engine, sel.selected_columns, chunk_size_frac=chunk_size_frac
        )
    else:
        chunk_size = in_clauses_chunk_size
    return (sel.where(in_col.in_(values)) for values in chunked(in_values, chunk_size))


__all__ = ["insert_dataframe", "select_to_dataframe"]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable, Iterable, Mapping, Sequence
from dataclasses import dataclass, is_dataclass
from functools import partial
from itertools import chain
from pathlib import Path
from re import search
from types import NoneType, UnionType
from typing import (
    Any,
    ForwardRef,
    Literal,
    NamedTuple,
    NotRequired,
    Optional,  # pyright: ignore[reportDeprecated]
    TypeAliasType,
    TypeGuard,
    Union,  # pyright: ignore[reportDeprecated]
    _TypedDictMeta,  # pyright: ignore[reportAttributeAccessIssue]
    cast,
    get_origin,
    overload,
    override,
)
from typing import get_args as _get_args
from typing import get_type_hints as _get_type_hints
from uuid import UUID
from warnings import warn

import whenever
from whenever import (
    Date,
    DateDelta,
    DateTimeDelta,
    PlainDateTime,
    Time,
    TimeDelta,
    ZonedDateTime,
)

from utilities.iterables import unique_everseen
from utilities.sentinel import Sentinel
from utilities.types import (
    Dataclass,
    StrMapping,
    StrStrMapping,
    TupleOrStrMapping,
    TypeLike,
)


def get_args(obj: Any, /, *, optional_drop_none: bool = False) -> tuple[Any, ...]:
    """Get the arguments of an annotation."""
    if isinstance(obj, TypeAliasType):
        return get_args(obj.__value__)
    args = _get_args(obj)
    if is_optional_type(obj) and optional_drop_none:
        args = tuple(a for a in args if a is not NoneType)
    return args


##


def get_forward_ref_args(obj: Any, /) -> StrStrMapping:
    """Get the forward args."""
    return {
        k: v.__forward_arg__
        for k, v in obj.__annotations__.items()
        if isinstance(v, ForwardRef)
    }


##


def get_literal_elements(obj: Any, /) -> list[Any]:
    """Get the elements of a literal annotation."""
    return _get_literal_elements_inner(obj)


def _get_literal_elements_inner(obj: Any, /) -> list[Any]:
    if isinstance(obj, str | int):
        return [obj]
    args = get_args(obj)
    parts = chain.from_iterable(map(_get_literal_elements_inner, args))
    return list(unique_everseen(parts))


##


def get_type_classes(obj: Any, /) -> tuple[type[Any], ...]:
    """Get the type classes from a type/tuple/Union type."""
    types: Sequence[type[Any]] = []
    if isinstance(obj, type):
        types.append(obj)
    elif isinstance(obj, tuple):
        for arg in obj:
            if isinstance(arg, type):
                types.append(arg)
            elif isinstance(arg, tuple):
                types.extend(get_type_classes(arg))
            elif is_union_type(arg):
                types.extend(get_union_type_classes(arg))
            else:
                raise _GetTypeClassesTupleError(obj=obj, inner=arg)
    elif is_union_type(obj):
        types.extend(get_union_type_classes(obj))
    else:
        raise _GetTypeClassesTypeError(obj=obj)
    return tuple(types)


@dataclass(kw_only=True, slots=True)
class GetTypeClassesError(Exception):
    obj: Any


@dataclass(kw_only=True, slots=True)
class _GetTypeClassesTypeError(GetTypeClassesError):
    @override
    def __str__(self) -> str:
        return f"Object must be a type, tuple or Union type; got {self.obj!r} of type {type(self.obj)!r}"


@dataclass(kw_only=True, slots=True)
class _GetTypeClassesTupleError(GetTypeClassesError):
    inner: Any

    @override
    def __str__(self) -> str:
        return f"Tuple must contain types, tuples or Union types only; got {self.inner} of type {type(self.inner)!r}"


##


def get_type_hints(
    obj: Any,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> dict[str, Any]:
    """Get the type hints of an object."""
    _ = {
        Date,
        DateDelta,
        DateTimeDelta,
        Literal,
        Path,
        PlainDateTime,
        Sentinel,
        StrMapping,
        Time,
        TimeDelta,
        UUID,
        ZonedDateTime,
        whenever.Date,
        whenever.DateDelta,
        whenever.DateTimeDelta,
        whenever.PlainDateTime,
        whenever.Time,
        whenever.TimeDelta,
        whenever.ZonedDateTime,
    }
    globalns_use = globals() | ({} if globalns is None else dict(globalns))
    localns_use = {} if localns is None else dict(localns)
    result: dict[str, Any] = obj.__annotations__
    result = result | dict(get_forward_ref_args(obj))
    try:
        hints = _get_type_hints(
            obj, globalns=globalns_use, localns=localns_use, include_extras=True
        )
    except NameError as error:
        if warn_name_errors:
            warn(f"Error getting type hints for {obj!r}; {error}", stacklevel=2)
    else:
        result = result | {
            key: value
            for key, value in hints.items()
            if (key not in result) or isinstance(result[key], str)
        }
    return result


##


def get_union_type_classes(obj: Any, /) -> tuple[type[Any], ...]:
    """Get the type classes from a Union type."""
    if not is_union_type(obj):
        raise _GetUnionTypeClassesUnionTypeError(obj=obj)
    types_: Sequence[type[Any]] = []
    for arg in get_args(obj):
        if not isinstance(arg, type):
            raise _GetUnionTypeClassesInternalTypeError(obj=obj, inner=arg)
        types_.append(arg)
    return tuple(types_)


@dataclass(kw_only=True, slots=True)
class GetUnionTypeClassesError(Exception):
    obj: Any


@dataclass(kw_only=True, slots=True)
class _GetUnionTypeClassesUnionTypeError(GetUnionTypeClassesError):
    @override
    def __str__(self) -> str:
        return (
            f"Object must be a Union type; got {self.obj!r} of type {type(self.obj)!r}"
        )


@dataclass(kw_only=True, slots=True)
class _GetUnionTypeClassesInternalTypeError(GetUnionTypeClassesError):
    inner: Any

    @override
    def __str__(self) -> str:
        return f"Union type must contain types only; got {self.inner} of type {type(self.inner)!r}"


##


def is_dataclass_class(obj: Any, /) -> TypeGuard[type[Dataclass]]:
    """Check if an object is a dataclass."""
    return isinstance(obj, type) and is_dataclass(obj)


##


def is_dataclass_instance(obj: Any, /) -> TypeGuard[Dataclass]:
    """Check if an object is an instance of a dataclass."""
    return (not isinstance(obj, type)) and is_dataclass(obj)


##


def is_dict_type(obj: Any, /) -> bool:
    """Check if an object is a dict type annotation."""
    return _is_annotation_of_type(obj, dict)


##


def is_frozenset_type(obj: Any, /) -> bool:
    """Check if an object is a frozenset type annotation."""
    return _is_annotation_of_type(obj, frozenset)


##


@overload
def is_instance_gen[T](
    obj: Any,
    type_: type[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[T]: ...
@overload
def is_instance_gen[T1](
    obj: Any,
    type_: tuple[T1],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[T1]: ...
@overload
def is_instance_gen[T1, T2](
    obj: Any,
    type_: tuple[T1, T2],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[T1 | T2]: ...
@overload
def is_instance_gen[T1, T2, T3](
    obj: Any,
    type_: tuple[T1, T2, T3],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[T1 | T2 | T3]: ...
@overload
def is_instance_gen[T1, T2, T3, T4](
    obj: Any,
    type_: tuple[T1, T2, T3, T4],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[T1 | T2 | T3 | T4]: ...
@overload
def is_instance_gen[T1, T2, T3, T4, T5](
    obj: Any,
    type_: tuple[T1, T2, T3, T4, T5],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[T1 | T2 | T3 | T4 | T5]: ...
@overload
def is_instance_gen(
    obj: Any,
    type_: Any,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> bool: ...
def is_instance_gen(
    obj: Any,
    type_: Any,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> bool:
    """Check if an instance relationship holds, except bool<int."""
    # parent
    if isinstance(type_, tuple):
        return any(
            is_instance_gen(
                obj,
                t,
                globalns=globalns,
                localns=localns,
                warn_name_errors=warn_name_errors,
            )
            for t in type_
        )
    if is_literal_type(type_):
        return obj in get_args(type_)
    if is_union_type(type_):
        return any(
            is_instance_gen(
                obj,
                t,
                globalns=globalns,
                localns=localns,
                warn_name_errors=warn_name_errors,
            )
            for t in get_args(type_)
        )
    # tuple vs tuple
    if isinstance(obj, tuple) and is_tuple_type(type_):
        type_args = get_args(type_)
        return (len(obj) == len(type_args)) and all(
            is_instance_gen(
                o,
                t,
                globalns=globalns,
                localns=localns,
                warn_name_errors=warn_name_errors,
            )
            for o, t in zip(obj, type_args, strict=True)
        )
    if isinstance(obj, tuple) is not is_tuple_type(type_):
        return False
    # basic
    if isinstance(type_, _TypedDictMeta):
        return _is_instance_typed_dict(
            obj,
            type_,
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
        )
    if isinstance(type_, type):
        return any(_is_instance_gen_type(obj, t) for t in get_type_classes(type_))
    raise IsInstanceGenError(obj=obj, type_=type_)


def _is_instance_typed_dict[T: _TypedDictMeta](
    obj: Any,
    type_: type[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[T]:
    if not isinstance(obj, dict):
        return False
    if not all(isinstance(k, str) for k in obj):
        return False
    obj = cast("dict[str, Any]", obj)
    hints = get_type_hints(
        type_, globalns=globalns, localns=localns, warn_name_errors=warn_name_errors
    )
    optional = {
        k for k, v in type_.__annotations__.items() if is_not_required_annotation(v)
    }
    required = {k: v for k, v in hints.items() if k not in optional}
    if not set(obj).issuperset(required):
        return False
    return all(
        is_instance_gen(
            obj[k],
            required[k],
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
        )
        for k in required
    )


def _is_instance_gen_type[T](obj: Any, type_: type[T], /) -> TypeGuard[T]:
    return (
        isinstance(obj, type_)
        and not (
            isinstance(obj, bool)
            and issubclass(type_, int)
            and not issubclass(type_, bool)
        )
        and not (
            isinstance(obj, dt.datetime)
            and issubclass(type_, dt.date)
            and not issubclass(type_, dt.datetime)
        )
    )


@dataclass(kw_only=True, slots=True)
class IsInstanceGenError(Exception):
    obj: Any
    type_: Any

    @override
    def __str__(self) -> str:
        return f"Invalid arguments; got {self.obj!r} of type {type(self.obj)!r} and {self.type_!r} of type {type(self.type_)!r}"


##


@overload
def is_iterable_of[T](
    obj: Any,
    cls: type[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T]]: ...
@overload
def is_iterable_of[T1](
    obj: Any,
    cls: tuple[type[T1]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T1]]: ...
@overload
def is_iterable_of[T1, T2](
    obj: Any,
    cls: tuple[type[T1], type[T2]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T1 | T2]]: ...
@overload
def is_iterable_of[T1, T2, T3](
    obj: Any,
    cls: tuple[type[T1], type[T2], type[T3]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T1 | T2 | T3]]: ...
@overload
def is_iterable_of[T1, T2, T3, T4](
    obj: Any,
    cls: tuple[type[T1], type[T2], type[T3], type[T4]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T1 | T2 | T3 | T4]]: ...
@overload
def is_iterable_of[T1, T2, T3, T4, T5](
    obj: Any,
    cls: tuple[type[T1], type[T2], type[T3], type[T4], type[T5]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T1 | T2 | T3 | T4 | T5]]: ...
@overload
def is_iterable_of[T](
    obj: Any,
    cls: TypeLike[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T]]: ...
def is_iterable_of[T](
    obj: Any,
    cls: TypeLike[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Iterable[T]]:
    """Check if an object is a iterable of tuple or string mappings."""
    return isinstance(obj, Iterable) and all(
        is_instance_gen(
            o,
            cls,
            globalns=globalns,
            localns=localns,
            warn_name_errors=warn_name_errors,
        )
        for o in obj
    )


##


def is_list_type(obj: Any, /) -> bool:
    """Check if an object is a list type annotation."""
    return _is_annotation_of_type(obj, list)


##


def is_literal_type(obj: Any, /) -> bool:
    """Check if an object is a literal type annotation."""
    return _is_annotation_of_type(obj, Literal)


##


def is_mapping_type(obj: Any, /) -> bool:
    """Check if an object is a mapping type annotation."""
    return _is_annotation_of_type(obj, Mapping)


##


def is_namedtuple_class(obj: Any, /) -> TypeGuard[type[Any]]:
    """Check if an object is a namedtuple."""
    return isinstance(obj, type) and _is_namedtuple_core(obj)


def is_namedtuple_instance(obj: Any, /) -> bool:
    """Check if an object is an instance of a dataclass."""
    return (not isinstance(obj, type)) and _is_namedtuple_core(obj)


def _is_namedtuple_core(obj: Any, /) -> bool:
    """Check if an object is an instance of a dataclass."""
    try:
        (base,) = obj.__orig_bases__
    except (AttributeError, ValueError):
        return False
    return base is NamedTuple


##


def is_not_required_annotation(obj: Any) -> bool:
    """Check if an annotation is not required."""
    if is_not_required_type(obj):
        return True
    match obj:
        case str() as text:
            return bool(search("NotRequired", text))
        case ForwardRef() as fr:
            return is_not_required_annotation(fr.__forward_arg__)
        case _:
            return False


##


def is_not_required_type(obj: Any, /) -> bool:
    """Check if an object is a not-required type annotation."""
    return (obj is NotRequired) or _is_annotation_of_type(obj, NotRequired)


##


def is_optional_type(obj: Any, /) -> bool:
    """Check if an object is an optional type annotation."""
    is_optional = _is_annotation_of_type(obj, Optional)  # pyright: ignore[reportDeprecated]
    return is_optional or (
        is_union_type(obj) and any(a is NoneType for a in _get_args(obj))
    )


##


@overload
def is_sequence_of[T](
    obj: Any,
    cls: type[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T]]: ...
@overload
def is_sequence_of[T1](
    obj: Any,
    cls: tuple[type[T1]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T1]]: ...
@overload
def is_sequence_of[T1, T2](
    obj: Any,
    cls: tuple[type[T1], type[T2]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T1 | T2]]: ...
@overload
def is_sequence_of[T1, T2, T3](
    obj: Any,
    cls: tuple[type[T1], type[T2], type[T3]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T1 | T2 | T3]]: ...
@overload
def is_sequence_of[T1, T2, T3, T4](
    obj: Any,
    cls: tuple[type[T1], type[T2], type[T3], type[T4]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T1 | T2 | T3 | T4]]: ...
@overload
def is_sequence_of[T1, T2, T3, T4, T5](
    obj: Any,
    cls: tuple[type[T1], type[T2], type[T3], type[T4], type[T5]],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T1 | T2 | T3 | T4 | T5]]: ...
@overload
def is_sequence_of[T](
    obj: Any,
    cls: TypeLike[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T]]: ...
def is_sequence_of[T](
    obj: Any,
    cls: TypeLike[T],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> TypeGuard[Sequence[T]]:
    """Check if an object is a sequence of tuple or string mappings."""
    return isinstance(obj, Sequence) and is_iterable_of(
        obj, cls, globalns=globalns, localns=localns, warn_name_errors=warn_name_errors
    )


##


def is_sequence_of_tuple_or_str_mapping(
    obj: Any, /
) -> TypeGuard[Sequence[TupleOrStrMapping]]:
    """Check if an object is a sequence of tuple or string mappings."""
    return isinstance(obj, Sequence) and all(map(is_tuple_or_str_mapping, obj))


##


def is_sequence_type(obj: Any, /) -> bool:
    """Check if an object is a sequence type annotation."""
    return _is_annotation_of_type(obj, Sequence)


##


def is_set_type(obj: Any, /) -> bool:
    """Check if an object is a set type annotation."""
    return _is_annotation_of_type(obj, set)


##


def is_string_mapping(obj: Any, /) -> TypeGuard[StrMapping]:
    """Check if an object is a string mapping."""
    return isinstance(obj, Mapping) and is_iterable_of(obj, str)


##


@overload
def is_subclass_gen[T](cls: type[Any], parent: type[T], /) -> TypeGuard[type[T]]: ...
@overload
def is_subclass_gen[T1](
    cls: type[Any], parent: tuple[type[T1]], /
) -> TypeGuard[type[T1]]: ...
@overload
def is_subclass_gen[T1, T2](
    cls: type[Any], parent: tuple[type[T1], type[T2]], /
) -> TypeGuard[type[T1 | T2]]: ...
@overload
def is_subclass_gen[T1, T2, T3](
    cls: type[Any], parent: tuple[type[T1], type[T2], type[T3]], /
) -> TypeGuard[type[T1 | T2 | T3]]: ...
@overload
def is_subclass_gen[T1, T2, T3, T4](
    cls: type[Any], parent: tuple[type[T1], type[T2], type[T3], type[T4]], /
) -> TypeGuard[type[T1 | T2 | T3 | T4]]: ...
@overload
def is_subclass_gen[T1, T2, T3, T4, T5](
    cls: type[Any], parent: tuple[type[T1], type[T2], type[T3], type[T4], type[T5]], /
) -> TypeGuard[type[T1 | T2 | T3 | T4 | T5]]: ...
@overload
def is_subclass_gen(cls: Any, parent: Any, /) -> bool: ...
def is_subclass_gen(cls: Any, parent: Any, /) -> bool:
    """Generalized `issubclass`."""
    # child
    if isinstance(cls, tuple):
        return all(is_subclass_gen(c, parent) for c in cls)
    if is_literal_type(cls):
        types = tuple(map(type, get_args(cls)))
        return (
            is_literal_type(parent) and set(get_args(cls)).issubset(get_args(parent))
        ) or is_subclass_gen(types, parent)
    if is_union_type(cls):
        return all(is_subclass_gen(c, parent) for c in get_args(cls))
    # parent
    if isinstance(parent, tuple):
        return any(is_subclass_gen(cls, p) for p in parent)
    if is_literal_type(parent):
        return is_literal_type(cls) and set(get_args(cls)).issubset(get_args(parent))
    if is_union_type(parent):
        return any(is_subclass_gen(cls, p) for p in get_args(parent))
    # tuple vs tuple
    if is_tuple_type(cls) and is_tuple_type(parent):
        cls_args, parent_args = get_args(cls), get_args(parent)
        return (len(cls_args) == len(parent_args)) and all(
            is_subclass_gen(c, p) for c, p in zip(cls_args, parent_args, strict=True)
        )
    if is_tuple_type(cls) is not is_tuple_type(parent):
        return False
    # basic
    if isinstance(cls, type):
        return any(_is_subclass_gen_type(cls, p) for p in get_type_classes(parent))
    raise IsSubclassGenError(cls=cls)


def _is_subclass_gen_type[T](cls: type[Any], parent: type[T], /) -> TypeGuard[type[T]]:
    return (
        issubclass(cls, parent)
        and not (
            issubclass(cls, bool)
            and issubclass(parent, int)
            and not issubclass(parent, bool)
        )
        and not (
            issubclass(cls, dt.datetime)
            and issubclass(parent, dt.date)
            and not issubclass(parent, dt.datetime)
        )
    )


@dataclass(kw_only=True, slots=True)
class IsSubclassGenError(Exception):
    cls: Any

    @override
    def __str__(self) -> str:
        return f"Argument must be a class; got {self.cls!r} of type {type(self.cls)!r}"


##


def is_tuple(obj: Any, /) -> TypeGuard[tuple[Any, ...]]:
    """Check if an object is a tuple."""
    return isinstance(obj, tuple)


##


def is_tuple_or_str_mapping(obj: Any, /) -> TypeGuard[TupleOrStrMapping]:
    """Check if an object is a tuple or string mapping."""
    return is_tuple(obj) or is_string_mapping(obj)


##


def is_tuple_type(obj: Any, /) -> bool:
    """Check if an object is a tuple type annotation."""
    return _is_annotation_of_type(obj, tuple)


##


def is_union_type(obj: Any, /) -> bool:
    """Check if an object is a union type annotation."""
    is_old_union = _is_annotation_of_type(obj, Union)  # pyright: ignore[reportDeprecated]
    return is_old_union or _is_annotation_of_type(obj, UnionType)


##


def _is_annotation_of_type(obj: Any, origin: Any, /) -> bool:
    """Check if an object is an annotation with a given origin."""
    return (get_origin(obj) is origin) or (
        isinstance(obj, TypeAliasType) and _is_annotation_of_type(obj.__value__, origin)
    )


##


@overload
def make_isinstance[T](cls: type[T], /) -> Callable[[Any], TypeGuard[T]]: ...
@overload
def make_isinstance[T1](cls: tuple[type[T1]], /) -> Callable[[Any], TypeGuard[T1]]: ...
@overload
def make_isinstance[T1, T2](
    cls: tuple[type[T1], type[T2]], /
) -> Callable[[Any], TypeGuard[T1 | T2]]: ...
@overload
def make_isinstance[T1, T2, T3](
    cls: tuple[type[T1], type[T2], type[T3]], /
) -> Callable[[Any], TypeGuard[T1 | T2 | T3]]: ...
@overload
def make_isinstance[T1, T2, T3, T4](
    cls: tuple[type[T1], type[T2], type[T3], type[T4]], /
) -> Callable[[Any], TypeGuard[T1 | T2 | T3 | T4]]: ...
@overload
def make_isinstance[T1, T2, T3, T4, T5](
    cls: tuple[type[T1], type[T2], type[T3], type[T4], type[T5]], /
) -> Callable[[Any], TypeGuard[T1 | T2 | T3 | T4 | T5]]: ...
@overload
def make_isinstance[T](cls: TypeLike[T], /) -> Callable[[Any], TypeGuard[T]]: ...
def make_isinstance[T](cls: TypeLike[T], /) -> Callable[[Any], TypeGuard[T]]:
    """Make a curried `isinstance` function."""
    return partial(_make_instance_core, cls=cls)


def _make_instance_core[T](obj: Any, /, *, cls: TypeLike[T]) -> TypeGuard[T]:
    return is_instance_gen(obj, cls)


__all__ = [
    "GetTypeClassesError",
    "GetUnionTypeClassesError",
    "IsInstanceGenError",
    "IsSubclassGenError",
    "get_literal_elements",
    "get_type_classes",
    "get_type_hints",
    "get_union_type_classes",
    "is_dataclass_class",
    "is_dataclass_instance",
    "is_dict_type",
    "is_frozenset_type",
    "is_instance_gen",
    "is_iterable_of",
    "is_list_type",
    "is_literal_type",
    "is_mapping_type",
    "is_namedtuple_class",
    "is_namedtuple_instance",
    "is_not_required_annotation",
    "is_not_required_type",
    "is_optional_type",
    "is_sequence_of",
    "is_sequence_of_tuple_or_str_mapping",
    "is_sequence_type",
    "is_set_type",
    "is_string_mapping",
    "is_subclass_gen",
    "is_tuple",
    "is_tuple_or_str_mapping",
    "is_tuple_type",
    "is_union_type",
    "make_isinstance",
]



from __future__ import annotations

from zoneinfo import ZoneInfo

HongKong = ZoneInfo("Asia/Hong_Kong")
Tokyo = ZoneInfo("Asia/Tokyo")
USCentral = ZoneInfo("US/Central")
USEastern = ZoneInfo("US/Eastern")


__all__ = ["HongKong", "Tokyo", "USCentral", "USEastern"]



from __future__ import annotations

import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import override


def which(cmd: str, /) -> Path:
    path = shutil.which(cmd)
    if path is None:
        raise WhichError(cmd=cmd)
    return Path(path)


@dataclass(kw_only=True, slots=True)
class WhichError(Exception):
    cmd: str

    @override
    def __str__(self) -> str:
        return f"{self.cmd!r} not found"


__all__ = ["WhichError", "which"]
